/*****************************************************************
 * Copyright (c) 2017 Texas Instruments and others
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Paul Gingrichv - Initial API and implementation
 *****************************************************************/
var gc = gc || {};
gc.databind = gc.databind || {};
gc.databind.internal = gc.databind.internal || {};
gc.databind.internal.reg = gc.databind.internal.reg || {};

(function() // closure for private static methods and data.
{
	var REGISTER_MODEL = "reg";
	var REG_REFRESH_INTERVAL = "$refresh_interval";
	var REG_TARGET_CONNECTED = "$target_connected";
	
	gc.databind.internal.reg.RegisterModel = function(codec) 
	{
	    this._codec = codec;
        this._targetConnectedBind = new gc.databind.VariableBindValue(false, true);
        this._symbols = {};
        this._calculatedBindings = {};
	};
	
	gc.databind.internal.reg.RegisterModel.prototype = new gc.databind.AbstractBindFactory(REGISTER_MODEL);
	
	gc.databind.internal.reg.RegisterModel.prototype.createNewBind = function(uri)
	{
		if( uri === REG_REFRESH_INTERVAL)
		{
			return new gc.databind.RefreshIntervalBindValue();
		}
		else if (uri === REG_TARGET_CONNECTED)
		{
			return this._targetConnectedBind;
		}
		
		if (!this._refreshIntervalProvider)
		{
			this._refreshIntervalProvider = this.getBinding(REG_REFRESH_INTERVAL);
		}
		
		// support for .# for bits within fields or registers 
		var pos = uri.lastIndexOf('.');
		if (pos > 0)
		{
		    var bitNumber = uri.substring(pos+1);
		    if (!isNaN(bitNumber))
		    {
		        var parentBind = this.getBinding(uri.substring(0, pos));
		        if (parentBind)
		        {
		            // TODO: test that error messages in parent bind are propagated through the field bind
		            return new gc.databind.internal.reg.FieldBind(parentBind, +bitNumber);
		        }
		    }
		}
		
		var symbolData = this._symbols[uri];
		if (symbolData)
		{
		    if (symbolData.fullyQualifiedName !== REGISTER_MODEL + '.' + uri)
		    {
		        // make sure all aliases share the same fully qualified name binding.
		        return gc.databind.registry.getBinding(symbolData.fullyQualifiedName);
		    }
		    
		    if (symbolData.parentRegister)
		    {
		        // support for fields within registers
	            var registerBind = gc.databind.registry.getBinding(symbolData.parentRegister.fullyQualifiedName);
	            
                return new gc.databind.internal.reg.FieldBind(registerBind, symbolData.start, symbolData.stop);
		    }
		    
	        var result = new gc.databind.internal.reg.RegisterBind(uri, this, this._refreshIntervalProvider);
	        
	        // support for qualifiers in register symbol data
	        if (symbolData.type === 'R')
	        {
	            result.addQualifier('readonly');
	        }
	        else if (symbolData.type === 'W')
	        {
	            result.addQualifier('writeonly');
	        }
	        else if (symbolData.type === "const" || symbolData.type === "nonvolatile")
	        {
	            result.addQualifier(symbolData.type);
	        }
	        return result;
		}
		else if (this._calculatedBindings.hasOwnProperty(uri))
		{
		    var defaultModel = gc.databind.registry.defaultModel();
		    gc.databind.registry.defaultModel(REGISTER_MODEL);
		    var bindResult = gc.databind.registry.getBinding(this._calculatedBindings[uri]);
		    gc.databind.registry.defaultModel(defaultModel);
		    return bindResult;
		}
		else
		{
		    // symbolic information is not available 
            var message;
		    var segment = uri.split('.');
		    if (segment.length > 2)
		    {
		        message = 'Binding "' + uri + '" is not recognized.  Expecting to see <register name>.<bit field name>';
		    }
		    else if (!this._symbols[segment[0]])
		    {
		        message = 'Register "' + segment[0] + '" is not recognized.  Please check the spelling.'; 
		    }
		    else 
		    {
		        message = 'Bit field "' + segment[1] + '" is not recognized for register "' + segment[0] + '"';
		    }
            var errorBind = new gc.databind.VariableBindValue(undefined, true);
		    errorBind.setStatus(gc.databind.AbstractStatus.createErrorStatus(message));
		    return errorBind;
		}
	};
	
    gc.databind.internal.reg.RegisterModel.prototype.setConnectedState = function(newState)
    {
        this._targetConnectedBind.updateValue(newState);
    };
    
    gc.databind.internal.reg.RegisterModel.prototype.connect = function(dataValidationCallback)
    {
        this.callbackForDataValidation = dataValidationCallback;
        this.setConnectedState(true);
    };
    
    gc.databind.internal.reg.RegisterModel.prototype.disconnect = function()
    {
        this.callbackForDataValidation = undefined;
        this._encoder = undefined;
        this._decoder = undefined;
        this.setConnectedState(false);
    };

    gc.databind.internal.reg.RegisterModel.prototype.addSymbol = function(symbolName, symbolData)
    {
        if (this._symbols[symbolName] === undefined)
        {
            this._symbols[symbolName] = symbolData;
        }
        else
        {
            this._symbols[symbolName] = null; // remove duplicates from the symbol table.
        }
    };
    
    gc.databind.internal.reg.RegisterModel.prototype.addSymbols = function(registerData)
    {
        var devices = registerData.devices || [];
        for(var n = devices.length; n-- > 0; )
        {
            var device = devices[n];
            if (device.address) 
            {
                device.address = gc.utils.string2value(device.address);
            }

            var commInterface = device['interface'];
            if (!commInterface)
            {
                throw 'register-defs .json file is missing an interface field for device "' + device.name + '".';
            }
            if (!registerData[commInterface])
            {
                throw 'register-defs .json file is missing an entry for interface "' + commInterface + '".';
            }
            var groups = (device.data && device.data.regblocks) || [];
            for(var i = groups.length; i-- > 0; )
            {
                var symbolName;
                var group = groups[i];
                group.parentDevice = device;
                var regs = group.registers || [];
                for(var j = regs.length; j-- > 0; )
                {
                    var reg = regs[j];
                    reg.parentGroup = group;
                    reg.nBytes = Math.ceil(reg.size/8);
                    reg.addr = gc.utils.string2value(reg.addr);
                    var fields = reg.fields || [];
                    for(var k = fields.length; k-- > 0; )
                    {
                        var field = fields[k];
                        field.parentRegister = reg;
                        symbolName = field.name;
                        this.addSymbol(symbolName, field);
                        symbolName = reg.name + '.' + symbolName;
                        field.fullyQualifiedName = REGISTER_MODEL + '.' + symbolName;
                        this.addSymbol(symbolName, field);
                    }
                    symbolName = reg.name;
                    reg.fullyQualifiedName = REGISTER_MODEL + '.' + symbolName;
                    this.addSymbol(symbolName, reg);
                }
            }
            
            var bindings = device["calculated bindings"];
            if (bindings)
            {
                // copy bindings into calculated bindings
                for(var bindName in bindings)
                {
                    if (bindings.hasOwnProperty(bindName))
                    {
                        this._calculatedBindings[bindName] = bindings[bindName];
                    }
                }
            }
        }
    };
    
    gc.databind.internal.reg.RegisterModel.prototype.getSymbolSuggestions = function(prefix)
    {
        prefix = prefix || "";
        var result = [];
        for(var symbolName in this._symbols)
        {
            if (this._symbols.hasOwnProperty(symbolName) && symbolName.index(prefix) === 0)
            {
                result.push(symbolName);    
            }
        }
        return result;
    };
    
    gc.databind.internal.reg.RegisterModel.prototype.readValue = function(uri)
    {
        return this._codec.readValue(this._symbols[uri]);
    };
    
    gc.databind.internal.reg.RegisterModel.prototype.writeValue = function(uri, value)
    {
        return this._codec.writeValue(this._symbols[uri], value);
    };
    
    gc.databind.internal.reg.RegisterModel.prototype.isConnected = function()
    {
        return this._targetConnectedBind.getValue();
    };
}());






